/// <reference types="node" />
import { EventEmitter } from 'events';
import { U2FHIDDevice } from './u2f-hid-device';
export interface IU2FRegisterRequest {
    appId: string;
    challenge: string;
    version: string;
}
export interface IU2FRegisterResponse extends IU2FRegisterRequest {
    registrationData: string;
    clientData: string;
}
export interface IU2FAuthenticateRequest extends IU2FRegisterRequest {
    keyHandle: string;
}
export interface IU2FAuthenticateResponse extends IU2FAuthenticateRequest {
    signatureData: string;
    clientData: string;
}
export declare class U2FDevice extends EventEmitter {
    driver: U2FHIDDevice;
    interactionTimeout: number;
    interactionPollInterval: number;
    constructor(driver: U2FHIDDevice);
    version(): Promise<Buffer>;
    command(cmd: number, p1?: number, p2?: number, data?: Buffer): Promise<Buffer>;
    register(req: IU2FRegisterRequest): Promise<IU2FRegisterResponse>;
    authenticate(req: IU2FAuthenticateRequest): Promise<IU2FAuthenticateResponse>;
    checkOnly(req: IU2FAuthenticateRequest): Promise<boolean | undefined>;
    close(): void;
    private _sendCommand;
    private _onDisconnected;
}
