"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var events_1 = require("events");
var defer_1 = require("./defer");
var util_1 = require("./util");
// Raw U2F commands
var U2F_REGISTER = 0x01; // Registration command
var U2F_AUTHENTICATE = 0x02; // Authenticate/sign command
var U2F_VERSION = 0x03; // Read version string command
// const U2F_VENDOR_FIRST = 0xc0;
// const U2F_VENDOR_LAST = 0xff;
var U2F_AUTH_ENFORCE = 0x03; // Enforce user presence and sign
var U2F_AUTH_CHECK_ONLY = 0x07; // Check only
var U2F_CLIENT_DATA_TYP_REGISTER = 'navigator.id.finishEnrollment';
var U2F_CLIENT_DATA_TYP_AUTHENTICATE = 'navigator.id.getAssertion';
var ERROR_CODES = {
    SW_NO_ERROR: 0x9000,
    SW_WRONG_LENGTH: 0x6700,
    SW_CONDITIONS_NOT_SATISFIED: 0x6985,
    SW_WRONG_DATA: 0x6a80,
    SW_INS_NOT_SUPPORTED: 0x6d00,
};
var INVERT_ERROR_CODES = util_1.invert(ERROR_CODES);
// U2F Device raw interface.
// https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-raw-message-formats-v1.2-ps-20170411.html
var U2FDevice = /** @class */ (function (_super) {
    __extends(U2FDevice, _super);
    function U2FDevice(driver) {
        var _this = _super.call(this) || this;
        _this.driver = driver;
        _this.interactionTimeout = 30 * 1000;
        _this.interactionPollInterval = 200;
        _this._onDisconnected = function () {
            _this.emit('disconnected');
        };
        if (_this.driver.protocolVersion !== 2) {
            throw new Error("Driver has unsupported protocol version: " + _this.driver.protocolVersion + ". Only 2 suported.");
        }
        _this.driver.on('disconnected', _this._onDisconnected);
        return _this;
    }
    U2FDevice.prototype.version = function () {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.command(U2F_VERSION)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    // Send raw U2F Command using APDU message exchange protocol.
    // p1, p2 and data args are optional.
    U2FDevice.prototype.command = function (cmd, p1, p2, data) {
        if (p1 === void 0) { p1 = 0; }
        if (p2 === void 0) { p2 = 0; }
        if (data === void 0) { data = Buffer.alloc(0); }
        return __awaiter(this, void 0, void 0, function () {
            var buf, deferred;
            return __generator(this, function (_a) {
                buf = Buffer.alloc(data.length + 7);
                buf[0] = 0; // CLA
                buf[1] = cmd; // INS
                buf[2] = p1; // P1
                buf[3] = p2; // P2
                buf[4] = 0; // LC1 (MSB)
                buf.writeUInt16BE(data.length, 5); // LC2, LC3 (LSB)
                data.copy(buf, 7);
                deferred = new defer_1.default();
                this._sendCommand(buf, p1, Date.now(), true, deferred);
                return [2 /*return*/, deferred.promise];
            });
        });
    };
    U2FDevice.prototype.register = function (req) {
        return __awaiter(this, void 0, void 0, function () {
            var clientData, buf, data;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        clientData = JSON.stringify({
                            typ: U2F_CLIENT_DATA_TYP_REGISTER,
                            challenge: req.challenge,
                            origin: req.appId,
                        });
                        buf = Buffer.concat([util_1.hash(clientData), util_1.hash(req.appId)]);
                        return [4 /*yield*/, this.command(U2F_REGISTER, 0, 0, buf)];
                    case 1:
                        data = _a.sent();
                        return [2 /*return*/, __assign(__assign({}, req), { registrationData: util_1.toWebsafeBase64(data), clientData: util_1.toWebsafeBase64(clientData) })];
                }
            });
        });
    };
    U2FDevice.prototype.authenticate = function (req) {
        return __awaiter(this, void 0, void 0, function () {
            var clientData, keyHandle, len, buf, data;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        clientData = JSON.stringify({
                            typ: U2F_CLIENT_DATA_TYP_AUTHENTICATE,
                            challenge: req.challenge,
                            origin: req.appId,
                        });
                        keyHandle = Buffer.from(req.keyHandle, 'base64');
                        len = keyHandle.length;
                        buf = Buffer.concat([
                            util_1.hash(clientData),
                            util_1.hash(req.appId),
                            Buffer.alloc(1, len > 0xff ? 0 : len),
                            keyHandle,
                        ]);
                        return [4 /*yield*/, this.command(U2F_AUTHENTICATE, U2F_AUTH_ENFORCE, 0, buf)];
                    case 1:
                        data = _a.sent();
                        return [2 /*return*/, __assign(__assign({}, req), { signatureData: util_1.toWebsafeBase64(data), clientData: util_1.toWebsafeBase64(clientData) })];
                }
            });
        });
    };
    U2FDevice.prototype.checkOnly = function (req) {
        return __awaiter(this, void 0, void 0, function () {
            var clientData, keyHandle, buf, e_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        clientData = '';
                        keyHandle = Buffer.from(req.keyHandle, 'base64');
                        buf = Buffer.concat([
                            util_1.hash(clientData),
                            util_1.hash(req.appId),
                            Buffer.from([keyHandle.length]),
                            keyHandle,
                        ]);
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 3, , 4]);
                        return [4 /*yield*/, this.command(U2F_AUTHENTICATE, U2F_AUTH_CHECK_ONLY, 0, buf)];
                    case 2:
                        _a.sent();
                        return [3 /*break*/, 4];
                    case 3:
                        e_1 = _a.sent();
                        if (e_1.code === ERROR_CODES.SW_CONDITIONS_NOT_SATISFIED) {
                            return [2 /*return*/, true];
                        }
                        if (e_1.code === ERROR_CODES.SW_WRONG_DATA) {
                            return [2 /*return*/, false];
                        }
                        throw e_1;
                    case 4: return [2 /*return*/];
                }
            });
        });
    };
    U2FDevice.prototype.close = function () {
        this.driver.close();
    };
    U2FDevice.prototype._sendCommand = function (buf, p1, startTime, shouldSendUserPresenceEvent, deferred) {
        return __awaiter(this, void 0, void 0, function () {
            var res, status, message, err;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.driver.msg(buf)];
                    case 1:
                        res = _a.sent();
                        if (res.length < 2) {
                            throw new Error('Cannot decode APDU: returned data too short.');
                        }
                        status = res.readUInt16BE(res.length - 2);
                        if (status === ERROR_CODES.SW_NO_ERROR) {
                            // Success; return data
                            return [2 /*return*/, deferred.resolve(res.slice(0, -2))];
                        }
                        else if (status === ERROR_CODES.SW_CONDITIONS_NOT_SATISFIED &&
                            !(p1 & 0x04) &&
                            Date.now() - startTime < this.interactionTimeout) {
                            // We need user presence, but don't have it.
                            // Wink and retry.
                            if (shouldSendUserPresenceEvent) {
                                this.emit('user-presence-required');
                            }
                            this.driver.wink();
                            setTimeout(function () {
                                _this._sendCommand(buf, p1, startTime, false, deferred);
                            }, this.interactionPollInterval);
                        }
                        else {
                            message = INVERT_ERROR_CODES[status] || "SW_UNKNOWN_ERROR: 0x" + status.toString(16);
                            err = new Error(message);
                            err.code = status;
                            deferred.reject(err);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    return U2FDevice;
}(events_1.EventEmitter));
exports.U2FDevice = U2FDevice;
//# sourceMappingURL=u2f-device.js.map