"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var crypto_1 = require("crypto");
var events_1 = require("events");
var node_hid_1 = require("node-hid");
var defer_1 = require("./defer");
var util_1 = require("./util");
var U2FHID_PING = 0x80 | 0x01;
var U2FHID_MSG = 0x80 | 0x03;
// const U2FHID_LOCK = 0x80 | 0x04;
var U2FHID_INIT = 0x80 | 0x06;
var U2FHID_WINK = 0x80 | 0x08;
// const U2FHID_SYNC = 0x80 | 0x3c;
var U2FHID_ERROR = 0x80 | 0x3f;
// const U2FHID_VENDOR_FIRST = 0x80 | 0x40;
// const U2FHID_VENDOR_LAST = 0x80 | 0x7f;
var U2FHID_BROADCAST_CID = 0xffffffff;
var REPORT_SIZE = 64;
var HID_ERRORS = {
    0x00: 'No error',
    0x01: 'Invalid command',
    0x02: 'Invalid parameter',
    0x03: 'Invalid message length',
    0x04: 'Invalid message sequencing',
    0x05: 'Message has timed out',
    0x06: 'Channel busy',
    0x0a: 'Command requires channel lock',
    0x0b: 'SYNC command failed',
    0x7f: 'Other unspecified error',
};
// FIDO U2F HID Protocol Specification
// https://fidoalliance.org/specs/fido-u2f-v1.2-ps-20170411/fido-u2f-hid-protocol-v1.2-ps-20170411.html
var U2FHIDDevice = /** @class */ (function (_super) {
    __extends(U2FHIDDevice, _super);
    function U2FHIDDevice(deviceInfo) {
        var _this = _super.call(this) || this;
        _this.protocolVersion = 0;
        _this.deviceVersion = [0, 0, 0];
        _this.caps = {};
        _this.closed = false;
        _this._channelId = U2FHID_BROADCAST_CID;
        _this._packetBuf = Buffer.alloc(REPORT_SIZE);
        _this._transactionQueue = [];
        _this._onError = function () {
            _this.close();
            _this.emit('disconnected');
        };
        _this._onData = function (buf) {
            util_1.debug('ON DATA', buf.toString('hex'));
            // Ignore packets outside the transaction.
            if (!_this._curTransaction) {
                return;
            }
            // Decode packet
            var channelId = buf.readUInt32BE(0);
            // Ignore packet addressed to other channels.
            if (channelId !== _this._channelId) {
                return;
            }
            var cmd = buf.readUInt8(4);
            if (cmd === U2FHID_ERROR) {
                // Errored.
                var errCode = buf.readUInt8(7);
                var error = new Error(HID_ERRORS[errCode] || HID_ERRORS[0x7f]);
                error.code = errCode;
                _this._curTransaction.deferred.reject(error);
                return _this._executeNextTransaction();
            }
            else if (cmd & 0x80) {
                // console.log('xx', cmd, buf)
                // Initial packet
                if (cmd !== _this._curTransaction.command) {
                    return console.error('Transaction decoding failure: response is for different operation: ', cmd, _this._curTransaction);
                }
                _this._curTransaction.toReceive = buf.readUInt16BE(5);
                _this._curTransaction.receivedBufs[0] = buf.slice(7);
                _this._curTransaction.receivedBytes += _this._curTransaction.receivedBufs[0].length;
            }
            else {
                // Continuation packet.
                _this._curTransaction.receivedBufs[cmd + 1] = buf.slice(5);
                _this._curTransaction.receivedBytes += _this._curTransaction.receivedBufs[cmd + 1].length;
            }
            // Call callback and finish transaction if read fully.
            // console.log('receviedBytes, toReceive', t.receivedBytes, t.toReceive)
            if (_this._curTransaction.receivedBytes >= _this._curTransaction.toReceive) {
                // console.log('receivedBufs', t.receivedBufs)
                _this._curTransaction.deferred.resolve(Buffer.concat(_this._curTransaction.receivedBufs).slice(0, _this._curTransaction.toReceive));
                _this._executeNextTransaction();
            }
        };
        if (!deviceInfo.path) {
            throw new Error('Device path does not exist');
        }
        _this.device = new node_hid_1.HID(deviceInfo.path);
        _this.device.on('error', _this._onError);
        _this.device.on('data', _this._onData);
        return _this;
    }
    U2FHIDDevice.open = function (device) {
        return __awaiter(this, void 0, void 0, function () {
            var u2fHid;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        u2fHid = new U2FHIDDevice(device);
                        return [4 /*yield*/, u2fHid.init()];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    U2FHIDDevice.prototype.init = function (forSure) {
        if (forSure === void 0) { forSure = false; }
        return __awaiter(this, void 0, void 0, function () {
            var nonce, data, respNonce;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        nonce = crypto_1.pseudoRandomBytes(8);
                        return [4 /*yield*/, this.command(U2FHID_INIT, nonce)];
                    case 1:
                        data = _a.sent();
                        respNonce = data.slice(0, 8);
                        if (nonce.toString('hex') !== respNonce.toString('hex')) {
                            throw new Error('Error initializing U2F HID Device: incorrect nonce');
                        }
                        // Decode other initialization data.
                        // try {
                        this._channelId = data.readUInt32BE(8);
                        this.protocolVersion = data.readUInt8(12);
                        this.deviceVersion = [data.readUInt8(13), data.readUInt8(14), data.readUInt8(15)];
                        this.caps = {
                            wink: !!(data.readUInt8(16) & 0x01),
                        };
                        // } catch (e) {
                        //   throw new Error('Error initializing U2F HID device: returned initialization data too short.');
                        // }
                        // Check protocol version is compatible.
                        if (this.protocolVersion !== 2) {
                            throw new Error("Error initializing U2F HID device: incompatible protocol version: " + this.protocolVersion);
                        }
                        if (!(this._channelId >>> 24 === 0 && !forSure)) return [3 /*break*/, 3];
                        // Some buggy keys give unacceptable channel_ids the first
                        // time (which don't work for following commands), so we try again.
                        this._channelId = U2FHID_BROADCAST_CID;
                        return [4 /*yield*/, this.init(true)];
                    case 2: return [2 /*return*/, _a.sent()];
                    case 3: return [2 /*return*/, this];
                }
            });
        });
    };
    U2FHIDDevice.prototype.command = function (command, data) {
        if (data === void 0) { data = Buffer.alloc(0); }
        if (this.closed) {
            throw new Error('Sending command to a closed device');
        }
        var deferred = new defer_1.default();
        this._transactionQueue.push({
            command: command,
            data: data,
            toReceive: 0xffff,
            receivedBytes: 0,
            receivedBufs: [],
            deferred: deferred,
        });
        if (!this._curTransaction) {
            this._executeNextTransaction();
        }
        return deferred.promise;
    };
    U2FHIDDevice.prototype.close = function () {
        if (this.closed) {
            return;
        }
        this.closed = true;
        this._curTransaction = undefined;
        this._transactionQueue = [];
        this.device.close();
        this.emit('closed');
    };
    U2FHIDDevice.prototype.ping = function (data) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.command(U2FHID_PING, data)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    U2FHIDDevice.prototype.wink = function () {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        if (!this.caps.wink) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.command(U2FHID_WINK)];
                    case 1: return [2 /*return*/, _a.sent()];
                    case 2: return [2 /*return*/];
                }
            });
        });
    };
    U2FHIDDevice.prototype.msg = function (data) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.command(U2FHID_MSG, data)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    U2FHIDDevice.prototype._executeNextTransaction = function () {
        this._curTransaction = this._transactionQueue.shift();
        if (!this._curTransaction) {
            return;
        }
        try {
            this._sendCommand(this._curTransaction.command, this._curTransaction.data);
        }
        catch (e) {
            // Can be either incorrect command/data, or the device is failed/disconnected ("Cannot write to HID device").
            // In the latter case, an 'error' event will be emitted soon.
            // TODO: We're probably in an inconsistent state now. Maybe we need to U2FHID_SYNC.
            this._curTransaction.deferred.reject(e);
            this._executeNextTransaction(); // Process next one.
        }
    };
    U2FHIDDevice.prototype._sendCommand = function (command, data) {
        if (data === void 0) { data = Buffer.alloc(0); }
        if (!(0x80 <= command && command < 0x100)) {
            throw new Error('Tried to send incorrect U2F HID command: ' + command);
        }
        // Create & send initial packet.
        var buf = this._packetBuf;
        buf.fill(0);
        buf.writeUInt32BE(this._channelId, 0);
        buf.writeUInt8(command, 4);
        buf.writeUInt16BE(data.length, 5);
        data.copy(buf, 7);
        data = data.slice(buf.length - 7);
        var initData = Array.from(buf);
        initData.unshift(0);
        util_1.debug('WRITE INIT DATA', "00" + buf.toString('hex'));
        this.device.write(initData);
        // Create & send continuation packets.
        var seq = 0;
        while (data.length > 0 && seq < 0x80) {
            buf.fill(0);
            buf.writeUInt32BE(this._channelId, 0);
            buf.writeUInt8(seq++, 4);
            data.copy(buf, 5);
            data = data.slice(buf.length - 5);
            var continueData = Array.from(buf);
            continueData.unshift(0);
            util_1.debug('WRITE CONTINUE DATA', "00" + buf.toString('hex'));
            this.device.write(continueData);
        }
        if (data.length > 0) {
            throw new Error("Tried to send too large data packet to U2F HID device (" + data.length + " bytes didn't fit).");
        }
    };
    return U2FHIDDevice;
}(events_1.EventEmitter));
exports.U2FHIDDevice = U2FHIDDevice;
//# sourceMappingURL=u2f-hid-device.js.map