"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var crypto_1 = require("crypto");
var node_hid_1 = require("node-hid");
var FIDO_USAGE_PAGE = 0xf1d0;
var FIDO_USAGE_U2FHID = 1;
function enumerateDevices(detectFn) {
    if (!detectFn) {
        detectFn = function (deviceInfo) {
            var isCompatible = deviceInfo.usagePage === FIDO_USAGE_PAGE && deviceInfo.usage === FIDO_USAGE_U2FHID;
            return isCompatible;
        };
    }
    return node_hid_1.devices().filter(detectFn);
}
exports.enumerateDevices = enumerateDevices;
function invert(obj) {
    var newObj = {};
    Object.keys(obj).forEach(function (key) {
        newObj[obj[key]] = key;
    });
    return newObj;
}
exports.invert = invert;
function toWebsafeBase64(buf) {
    if (!Buffer.isBuffer(buf)) {
        buf = Buffer.from(buf);
    }
    return buf
        .toString('base64')
        .replace(/\//g, '_')
        .replace(/\+/g, '-')
        .replace(/=+$/, '');
}
exports.toWebsafeBase64 = toWebsafeBase64;
function fromWebsafeBase64(base64) {
    var normalBase64 = base64.replace(/-/g, '+').replace(/_/g, '/') + '=='.substring(0, (3 * base64.length) % 4);
    return Buffer.from(normalBase64, 'base64');
}
exports.fromWebsafeBase64 = fromWebsafeBase64;
function hash(data) {
    return crypto_1.createHash('SHA256')
        .update(data)
        .digest();
}
exports.hash = hash;
function debug() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    if (process.env.DEBUG_DEVICE_DATA) {
        console.log.apply(console, args);
    }
}
exports.debug = debug;
function delay(ms) {
    return new Promise(function (resolve) { return setTimeout(resolve, ms); });
}
exports.delay = delay;
//# sourceMappingURL=util.js.map